!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!module computes flat sky transfer functions for B (from tensors) and T (from scalars)!!!!
!Can be modified to compute other transferf (with same input from CAMB) D. Meerburg 2015!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module FlatSkyTranfer

  use settings
  use interpolation

  implicit none


contains 

  subroutine Compute_TransferFlat

    integer     :: i, imax = 100000000, j, n, eof !imax has to be some large number to loop out the file
    real(dl),allocatable    :: SourcDatScal(:,:,:)
    real(dl),allocatable    :: SourcDatPolScal(:,:,:) 
    real(dl),allocatable    :: SourcDatPolTens(:,:,:) 
    real(dl),allocatable    :: karrScal(:), karrPolTens(:), karrPolScal(:) !number of tau samples for each k
    real(dl),allocatable    :: tauarrScal(:,:), tauarrPolTens(:,:), tauarrPolScal(:,:)
    real(dl)    :: k, tau, S1, S2, S3
    real(dl)    :: tempk, temptau
    integer     :: intkmaxScal, intkmaxPolTens, intkmaxPolScal, inttaumaxPolTens, inttaumaxPolScal, inttaumaxScal

    real(dl)    :: dkS, dkT, dtauS, dtauT, tauintS(4), tauintT(4), kintresS, kintresT  
    real(dl)    :: qS, qT, q2S, q2T, qS1, qS2 
    real(dl),allocatable :: sp1s(:), sp2s(:), sp3s(:) 	
    real(dl),allocatable :: SourceFScal(:,:,:), SourceFPolTens(:,:,:),  SourceFPolScal(:,:,:)
    real(dl),allocatable :: sp1t(:), sp2t(:), sp3t(:) 
    integer :: ell, s, smaxPoltens, smaxScal,  smaxPolScal
    real(dl) :: TTtensFac
    real(dl) :: qSlow, qTlow, klowS, klowT
    character(100) :: fnSLSS,fnPolTLSS,fnPolSLSS

    real(dl) :: tauIntegrandS(4), tauIntegrandT(4)
    real(dl) :: Bfac !for testing



    !required for limited stack size MacOsX + openmp
    !can improve here, since the first array can be merged.
    !Someone else can do that...y
    if(scalars) then 
       allocate(SourcDatScal(3,3*lmaxScal,2000))
       allocate(karrScal(3*lmaxScal))
       allocate(tauarrScal(3*lmaxScal,2000))

    endif
    if(polarization .and. tensors) then 
       allocate(SourcDatPolTens(3,3*lmaxPolTens,2000))
       allocate(karrPolTens(3*lmaxPolTens))
       allocate(tauarrPolTens(3*lmaxPolTens,2000))

    endif
    if(polarization .and. scalars) then 
       allocate(SourcDatPolScal(3,3*lmaxPolScal,2000))
       allocate(karrPolScal(3*lmaxPolScal))
       allocate(tauarrPolScal(3*lmaxPolScal,2000))

    endif

    Etau =  exp(-ReiTau)
    !exchanged indices for tau to check for speed. Does not make 
    !a diference so kept it as was for tensors. 

    if (AppLss) then 
       fnSLSS = Trim(homefolder)//Trim(NameTransferT)//Trim(Scaltxt)
       fnPolTLSS = Trim(homefolder)//Trim(NameTransferB)//Trim(PolTtxt)
       fnPolSLSS = Trim(homefolder)//Trim(NameTransferE)//Trim(PolStxt)

    else
       fnSLSS = Trim(homefolder)//Trim(NameTransferT)//Trim('b')//Trim(Scaltxt)
       fnPolTLSS = Trim(homefolder)//Trim(NameTransferB)//Trim('b')//Trim(PolTtxt)
       fnPolSLSS = Trim(homefolder)//Trim(NameTransferE)//Trim('b')//Trim(PolStxt)
    endif
    
    write(*,*) ""
    if (CLASS) write(*,'(A30)') "Using CLASS Source functions  "
    if (CAMB) write(*,'(A30)') "Using CAMB Source functions   "
    write(*,*) ""

    if (scalars .and. Tmode) then 
       karrScal(1) = 0.d0
       j = 1
       n = 1
       tempk = karrScal(1)
       !if bispectrum, use SourcesS_BS.txt
       !else SourcesS.txt or any of the other ones
       fname = Trim(sourcefunctions)//Trim(NameSource)//Trim(Scaltxt)
       open(unit=21,file=fname, status='old')

       do i = 1, imax

          read(21,'(I5,4E16.7)',iostat=eof)s, k,tau,S1, S2
          !if (j .eq. 10) write(*,*) s, k, tau, S1, S2!, S3 
          if (eof .ne. 0 ) exit !end of file
          if (k > tempk) then
             !numtau(j) = n-1
             !write(*,*) "number of tau:", n-1
             j = j + 1
             n = 1
             !write(*,*) k			
          endif
          !write(*,*) i
          tempk = k
          karrScal(j) = k
          tauarrScal(j,n) = tau
          !needed for T and E
          SourcDatScal(1,j,n) = S1
          SourcDatScal(2,j,n) = S2
          !needed for B-modes
          !SourcDat(3,j,n) = S3
          n = n + 1		
       enddo
       intkmaxScal = j
       inttaumaxScal = n-1
       close(21)
       do i = 1, pstep

          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""
       if (AppLSS) write(*,'(A30)') "Thin Shell Approximation      "
       write(*,'(A30)') "<<<<<<Scalar-Intensity>>>>>>>>"
       write(*,*) ""
       write(*,'(A25,I5)') "Number of tau in scal:     ", inttaumaxScal
       write(*,'(A25,I5)') "Number of k in scal:       ", intkmaxScal
       if (do_log_ell) then
          write(*,'(A25)') "Doing log Grid in ell      "
       else
          write(*,'(A25,I5)') "Delta ell:                 ", lstep
       endif
       write(*,'(A25,F5.2)') "kmax:                      ", karrScal(intkmaxScal)
       write(*,'(A25,F8.2)') "Taumax:                    ", tauarrScal(inttaumaxScal,2)
       write(*,*) ""
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""

    endif
    
    
!!$    if (Tmode .and. scalars .and. CLASS) then 
!!$       karrScal(1) = 0.d0
!!$       j = 0
!!$       n = 2
!!$       temptau = 0.d0
!!$       !tempk = karrT(1)
!!$       !use Sources_BS.txt for bispectrum
!!$       fname = Trim(sourcefunctions)//Trim(NameSource)//Trim(Scaltxt)
!!$
!!$       open(unit=21,file=fname, status='old')
!!$       do i = 1, imax
!!$
!!$          read(21,*,iostat=eof) tau, k,S1
!!$          !if (j .eq. 10) write(*,*) s, k, tau, S1, S2, S3 
!!$          if (eof .ne. 0 ) exit !end of file
!!$          if (tau > temptau) then
!!$             !numtau(j) = n-1
!!$             !write(*,*) "number of tau:", n-1
!!$             j = j + 1
!!$             n = 2
!!$             !write(*,*) k			
!!$          endif
!!$          !write(*,*) i
!!$          temptau = tau
!!$          karrScal(n) = k
!!$          tauarrScal(n,j) = tau
!!$          !needed for T and E
!!$          SourcDatScal(1,n,j) = S1
!!$          !SourcDatT(2,j,n) = S2
!!$          !needed for B-modes
!!$          !SourcDatT(3,j,n) = S3
!!$          n = n + 1		
!!$       enddo
!!$       intkmaxScal =  n-1
!!$       inttaumaxScal = j
!!$       close(21)
!!$       !ell = 1000
!!$       do i = 1, pstep
!!$          write(*,'(a)',advance='no')  '*'
!!$       enddo
!!$       write(*,*) ""
!!$
!!$       if (AppLSS) write(*,'(A30)') "Thin Shell Approximation      "
!!$       write(*,'(A30)') "<<<<<<Scalar-Intensity>>>>>>>>"
!!$       write(*,*) ""
!!$       write(*,'(A25,I5)') "Number of tau in tens:     ", inttaumaxScal
!!$       write(*,'(A25,I5)') "Number of k in tens:       ", intkmaxScal
!!$       if (do_log_ell) then
!!$          write(*,'(A25)') "Doing log Grid in ell      "
!!$       else
!!$          write(*,'(A25,I5)') "Delta ell:                 ", lstep
!!$       endif
!!$       write(*,'(A25,F5.2)') "kmax:                      ", karrScal(intkmaxScal)
!!$       write(*,'(A25,F8.2)') "Taumax:                    ", tauarrScal(2,inttaumaxScal)
!!$       write(*,*) ""
!!$       do i = 1, pstep
!!$          write(*,'(a)',advance='no')  '*'
!!$       enddo
!!$       write(*,*) ""
!!$
!!$    endif

    if (polarization .and. tensors) then 
       karrPolTens(1) = 0.d0
       j = 0
       n = 2
       temptau = 0.d0
       !tempk = karrT(1)
       !use Sources_BS.txt for bispectrum
       fname = Trim(sourcefunctions)//Trim(NameSource)//Trim(PolTtxt)

       open(unit=21,file=fname, status='old')
       do i = 1, imax

          read(21,*,iostat=eof) tau, k,S1
          !if (j .eq. 10) write(*,*) s, k, tau, S1, S2, S3 
          if (eof .ne. 0 ) exit !end of file
          if (tau > temptau) then
             !numtau(j) = n-1
             !write(*,*) "number of tau:", n-1
             j = j + 1
             n = 2
             !write(*,*) k			
          endif
          !write(*,*) i
          temptau = tau
          karrPolTens(n) = k
          tauarrPolTens(n,j) = tau
          !write(*,*) n,j,tau
          !needed for T and E
          SourcDatPolTens(1,n,j) = S1
         
          !SourcDatT(2,j,n) = S2
          !needed for B-modes
          !SourcDatT(3,j,n) = S3
          n = n + 1		
       enddo
       intkmaxPolTens =  n-1
       inttaumaxPoltens = j
       close(21)
       !ell = 1000
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""

       if (AppLSS) write(*,'(A30)') "Thin Shell Approximation      "
       write(*,'(A30)') "<<<<<<Tensor-Polarization>>>>>"
       write(*,*) ""
       write(*,'(A25,I5)') "Number of tau in tens:     ", inttaumaxPolTens
       write(*,'(A25,I5)') "Number of k in tens:       ", intkmaxPolTens
       if (do_log_ell) then
          write(*,'(A25)') "Doing log Grid in ell      "
       else
          write(*,'(A25,I5)') "Delta ell:                 ", lstep
       endif
       write(*,'(A25,F5.2)') "kmax:                      ", karrPolTens(intkmaxPolTens)
       write(*,'(A25,F8.2)') "Taumax:                    ", tauarrPolTens(2,inttaumaxPolTens)
       write(*,*) ""
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""

    endif


    if (polarization .and. scalars) then 
       karrPolScal(1) = 0.d0
       j = 0
       n = 2
       temptau = 0.d0
       !tempk = karrT(1)
       !use Sources_BS.txt for bispectrum
       fname = Trim(sourcefunctions)//Trim(NameSource)//Trim(PolStxt)

       open(unit=21,file=fname, status='old')
       do i = 1, imax

          read(21,*,iostat=eof) tau, k,S1
          !if (j .eq. 10) write(*,*) s, k, tau, S1, S2, S3 
          if (eof .ne. 0 ) exit !end of file
          if (tau > temptau) then
             !numtau(j) = n-1
             !write(*,*) "number of tau:", n-1
             j = j + 1
             n = 2
             !write(*,*) k			
          endif
          !write(*,*) i
          temptau = tau
          karrPolScal(n) = k
          tauarrPolScal(n,j) = tau
          SourcDatPolScal(1,n,j) = S1
          !if(n .eq. 100) write(*,*) SourcDatPolScal(1,n,j), SourcDatScal(1,n,j)
          n = n + 1		
       enddo
       intkmaxPolScal =  n-1
       inttaumaxPolScal = j
       close(21)
       !ell = 1000
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""

       if (AppLSS) write(*,'(A30)') "Thin Shell Approximation      "
       write(*,'(A30)') "<<<<<<Scalar-Polarization>>>>>"
       write(*,*) ""
       write(*,'(A25,I5)') "Number of tau in scal(p):    ", inttaumaxPolScal
       write(*,'(A25,I5)') "Number of k in scal(p):      ", intkmaxPolScal
       if (do_log_ell) then
          write(*,'(A25)') "Doing log Grid in ell      "
       else
          write(*,'(A25,I5)') "Delta ell:                 ", lstep
       endif
       write(*,'(A25,F5.2)') "kmax:                      ", karrPolScal(intkmaxPolScal)
       write(*,'(A25,F8.2)') "Taumax:                    ", tauarrPolScal(2,inttaumaxPolScal)
       write(*,*) ""
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
    endif
    !turns out, you can get away with much fewer tau:
    !inttaumaxT = 500
    !can not get away with too few k however, especially if you do not use AppLSS = .true.
    
    if (do_log_ell) then
       smaxPolTens = 131 !(lmax = 500)
       smaxPolScal = 331 !(lmax = 4500)
       smaxScal = 331 !(lmax = 4500)   
    else
       smaxPolTens = int(lmaxPolTens/lstep)+1
       smaxPolScal = int(lmaxPolScal/lstep)+1
       smaxScal = int(lmaxScal/lstep)+1
    endif

    write(*,*) ""
    write(*,*) ""
    write(*,*) ""


    !precompute Source function fp
    !this is the slow part:
    if (polarization .and. tensors) then
       !for debugging
       !open(unit=28,file="/mnt/scratch-lustre/meerburg/SourceSplineTp2.txt", status="replace")
       i_progress = int(smaxPolTens/pstep)
       write(*,*) ""
       write(*,'(A55)') "Precomputing Spline Coefficients Tensors (Polarization)"
       do i = 1, pstep
          write(*,'(a)',advance='no')  '**'
       enddo
       write(*,*) ""
       allocate(sp1t(1:intkmaxPolTens))
       allocate(sp2t(1:intkmaxPoltens))
       allocate(sp3t(1:intkmaxPolTens))   
       allocate(SourceFPoltens(intkmaxPolTens,inttaumaxPoltens,smaxPoltens))
       !$OMP PARALLEL DO DEFAUlT(SHARED), &
       !$OMP PRIVATE(ell,j,i,s,sp1t,sp2t,sp3t,qT)

       !do ell = lmin, lmaxT, lstep
       do s = 1, smaxPoltens !s = ell
          if(do_log_ell) then
             ell = elltable(s)
          else            
             ell = lmin + lstep*(s-1)
          endif
          do j = 1,  inttaumaxPolTens !j = tau

             call spline(karrPolTens(2:intkmaxPolTens),SourcDatPolTens(1,2:intkmaxPolTens,j), &
                  sp1t,sp2t,sp3t,intkmaxPolTens-1)


             do i = 1, intkmaxPoltens !loop over k (inverted)
                qT = sqrt(karrPoltens(i)**2+(ell/(tauarrPolTens(i,j)-tau0))**2)

                SourceFPolTens(i,j,s) = & 
                     ispline(qT,karrPoltens(2:intkmaxPolTens),SourcDatPolTens(1,2:intkmaxPolTens,j),&
                     sp1t,sp2t,sp3t,intkmaxPolTens-1) 	
                !for debugging			
                !if(ell .eq. 90) write(28,*) ell, karrT(i), tauarrT(i,j), SourceFT(i,j,s)
             enddo
          enddo
          !write(*,*) ell
          if(mod(s,i_progress) .eq. 0) write(*,'(a)',advance='no')  '**'

       enddo
       !$OMP END PARALLEL DO 
       deallocate(sp1t)
       deallocate(sp2t)
       deallocate(sp3t)
       !close(28)
    endif
    if (polarization .and. scalars) then
       !for debugging
       !open(unit=28,file="/mnt/scratch-lustre/meerburg/SourceSplineTp2.txt", status="replace")
       i_progress = int(smaxPolScal/pstep)
       write(*,*) ""
       write(*,'(A55)') "Precomputing Spline Coefficients Scalars (Polarization)"
       do i = 1, pstep
          write(*,'(a)',advance='no')  '**'
       enddo
       write(*,*) ""
       if(CLASS) then
          allocate(sp1s(1:intkmaxPolScal))
          allocate(sp2s(1:intkmaxPolScal))
          allocate(sp3s(1:intkmaxPolScal))   
          allocate(SourceFPolScal(intkmaxPolScal,inttaumaxPolScal,smaxPolScal))
       elseif(CAMB) then
          allocate(sp1s(1:intkmaxScal))
          allocate(sp2s(1:intkmaxScal))
          allocate(sp3s(1:intkmaxScal))   
          allocate(SourceFPolScal(intkmaxScal,inttaumaxScal,smaxScal))
       endif
       !$OMP PARALLEL DO DEFAUlT(SHARED), &
       !$OMP PRIVATE(ell,j,i,s,sp1s,sp2s,sp3s,qT)

       !do ell = lmin, lmaxT, lstep
       do s = 1, smaxPolScal !s = ell
          if(do_log_ell) then
             ell = elltable(s)
          else            
             ell = lmin + lstep*(s-1)
          endif
          do j = 1,  inttaumaxPolScal !j = tau
             if(CLASS) then
                call spline(karrPolScal(2:intkmaxPolScal),SourcDatPolScal(1,2:intkmaxPolScal,j), &
                     sp1s,sp2s,sp3s,intkmaxPolScal-1)
             elseif(CAMB) then !using CAMB source functions
                call spline(karrScal(2:intkmaxScal),SourcDatScal(2,2:intkmaxScal,j), &
                     sp1s,sp2s,sp3s,intkmaxScal-1)
             endif


             do i = 1, intkmaxPolScal !loop over k (inverted)

                if(CLASS) then !using CLASS source fuctions
                   qT = sqrt(karrPolScal(i)**2+(ell/(tauarrPolScal(i,j)-tau0))**2)
                   SourceFPolScal(i,j,s) = & 
                        ispline(qT,karrPolScal(2:intkmaxPolScal),SourcDatPolScal(1,2:intkmaxPolScal,j), &
                        sp1s,sp2s,sp3s,intkmaxPolScal-1)
                elseif(CAMB) then !using CAMB source functions
                   qS = sqrt(karrScal(i)**2+(ell/(tauarrScal(i,j)-tau0))**2)
                   SourceFPolScal(i,j,s) = & 
                        ispline(qS,karrScal(2:intkmaxScal),SourcDatScal(2,2:intkmaxScal,j), &
                        sp1s,sp2s,sp3s,intkmaxScal-1)
                endif

                !for debugging			
                !if(ell .eq. 90) write(28,*) ell, karrT(i), tauarrT(i,j), SourceFT(i,j,s)
             enddo
          enddo
          !write(*,*) ell
          if(mod(s,i_progress) .eq. 0) write(*,'(a)',advance='no')  '**'

       enddo
       !$OMP END PARALLEL DO 
       deallocate(sp1s)
       deallocate(sp2s)
       deallocate(sp3s)
       !close(28)
    endif

    if (scalars .and. Tmode) then
       i_progress = int(smaxScal/pstep)
       write(*,*) ""
       !inttaumaxScal = 300
       write(*,'(A40)') "Precomputing Spline Coefficients Scalars"
       do i = 1, pstep
          write(*,'(a)',advance='no')  '**'
       enddo
       write(*,*) ""
       allocate(sp1s(1:intkmaxScal))
       allocate(sp2s(1:intkmaxScal))
       allocate(sp3s(1:intkmaxScal))
       allocate(SourceFScal(intkmaxScal,inttaumaxScal,smaxScal))
       !for debugging
       !open(unit=21,file="/mnt/scratch-lustre/meerburg/SourceSplineS.txt", status="replace")
       !$OMP PARALLEL DO DEFAUlT(SHARED), &
       !$OMP PRIVATE(ell,j,i,s,sp1s,sp2s,sp3s,qS)
       do s = 1, smaxScal
          if(do_log_ell) then
             ell = elltable(s)
          else            
             ell = lmin + lstep*(s-1)
          endif
          do j = 1,  inttaumaxScal 
             call spline(karrScal(2:intkmaxScal),SourcDatScal(1,2:intkmaxScal,j), &
                  sp1s,sp2s,sp3s,intkmaxScal-1)

             do i = 1, intkmaxScal !loop over k (inverted)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                !This is for T-modes. Change to columns 2 for E
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                qS = sqrt(karrScal(i)**2+(ell/(tauarrScal(i,j)-tau0))**2)
                SourceFScal(i,j,s) = & 
                     ispline(qS,karrScal(2:intkmaxScal),SourcDatScal(1,2:intkmaxScal,j), &
                     sp1s,sp2s,sp3s,intkmaxScal-1)
                !for debugging		
                !if(ell .eq. 100) write(21,*) ell, karrS(i), tauarrS(i,j), SourceFS(i,j,s)
             enddo
          enddo
          !write(*,*) ell
          if(mod(s,i_progress) .eq. 0) write(*,'(a)',advance='no')  '**'
       enddo
       !$OMP END PARALLEL DO 
       deallocate(sp1s)
       deallocate(sp2s)
       deallocate(sp3s)
       !close(21)
    endif


    if (scalars .and. Tmode) then
       open(unit=21,file=fnSLSS, status="replace")
       do s = 1, smaxScal !loop over ell
          if(do_log_ell) then
             ell = elltable(s)
          else            
             ell = lmin + lstep*(s-1)
          endif

          !flat sky TT 
          do i = 1, intkmaxScal !Integrate over k
             if (i < intkmaxScal) then 
                dkS = karrScal(i+1) - karrScal(i)
             else 
                dkS = karrScal(i) - karrScal(i-1)
             endif
             tauintS(1:4) = 0.d0
             do j = 1,  inttaumaxScal-1 !integration over tau
                if (i .eq. 1) then 
                   dtauS = (tauarrScal(2,j+1)-tauarrScal(2,j))
                else
                   dtauS = (tauarrScal(i,j+1)-tauarrScal(i,j))
                endif
                !dtauS = tauarrS(j,i)
                !two different integrals, one with sin, one with cos
                qS = sqrt(karrScal(i)**2+(ell/(tauarrScal(i,j)-tau0))**2)

	        !using approximation here by replacing tau0->tauR
	        !then also the spectrum should be computed with replacement. 
	        if(AppLSS) then
!!$                   TauIntegrandS(1) = (sin(karrScal(i)*(tauarrScal(i,j)-tauRS))*SourceFScal(i,j,s)/(tauarrScal(i,j)-tau0)**2 + &
!!$                        sin(karrScal(i)*(tauarrScal(i,j+1)-tauRS))*SourceFScal(i,j+1,s)/(tauarrScal(i,j+1)-tau0)**2)/2.0d0
!!$                   tauintS(1) = tauintS(1) + TauIntegrandS(1)*dtauS
!!$                   TauIntegrandS(2) = (cos(karrScal(i)*(tauarrScal(i,j)-tauRS))*SourceFScal(i,j,s)/(tauarrScal(i,j)-tau0)**2 + &
!!$                        cos(karrScal(i)*(tauarrScal(i,j+1)-tauRS))*SourceFScal(i,j+1,s)/(tauarrScal(i,j+1)-tau0)**2)/2.0d0
!!$                   tauintS(2) = tauintS(2) + TauIntegrandS(2)*dtauS
                   !if (karrS(i) .eq. 0) write(*,*) karrS(i), tauintS(2), SourceFS(i,j,s), TauIntegrandS(2)
                   tauintS(1) = tauintS(1) + sin(karrScal(i)*(tauarrScal(i,j)-tauRS))*SourceFScal(i,j,s) & 
                        *dtauS/(tauarrScal(i,j)-tau0)**2 !4 * (ell**2/qS**2) 
                   tauintS(2) = tauintS(2) + cos(karrScal(i)*(tauarrScal(i,j)-tauRS))*SourceFScal(i,j,s) & 
                        *dtauS/(tauarrScal(i,j)-tau0)**2 !4 * (ell**2/qS**2)
                else
                   !we need to compute the primordial power spectrum with k-par = l^2/D^2
                   qS1 = sqrt(karrScal(i)**2+(ell/(tauarrScal(i,j)-tau0))**2)
                   qS2 = sqrt(karrScal(i)**2+(ell/(tauarrScal(i,j+1)-tau0))**2)
                   !multiplied with D_1^2 and divided by k^3
                   TauIntegrandS(1) = (sin(karrScal(i)*(tauarrScal(i,j)-tauRS))*SourceFScal(i,j,s)/qs1**3 + &
                        sin(karrScal(i)*(tauarrScal(i,j+1)-tauRS))*SourceFScal(i,j+1,s)/qs2**3)/2.0d0
                   tauintS(1) = tauintS(1) + TauIntegrandS(1)*dtauS
                   TauIntegrandS(2) = (cos(karrScal(i)*(tauarrScal(i,j)-tauRS))*SourceFScal(i,j,s)/qs1**3 + &
                        cos(karrScal(i)*(tauarrScal(i,j+1)-tauRS))*SourceFScal(i,j+1,s)/qs2**3)/2.0d0
                   tauintS(2) = tauintS(2) + TauIntegrandS(2)*dtauS

                   !same as AppLSS
                   TauIntegrandS(3) = (sin(karrScal(i)*(tauarrScal(i,j)-tauRS))*SourceFScal(i,j,s)/(tauarrScal(i,j)-tau0)**2 + &
                        sin(karrScal(i)*(tauarrScal(i,j+1)-tauRS))*SourceFScal(i,j+1,s)/(tauarrScal(i,j+1)-tau0)**2)/2.0d0
                   tauintS(3) = tauintS(3) + TauIntegrandS(3)*dtauS
                   TauIntegrandS(4) = (cos(karrScal(i)*(tauarrScal(i,j)-tauRS))*SourceFScal(i,j,s)/(tauarrScal(i,j)-tau0)**2 + &
                        cos(karrScal(i)*(tauarrScal(i,j+1)-tauRS))*SourceFScal(i,j+1,s)/(tauarrScal(i,j+1)-tau0)**2)/2.0d0
                   tauintS(4) = tauintS(4) + TauIntegrandS(4)*dtauS
                endif

             enddo


             if (AppLSS) then 
                write(21,'(I4,3E17.7)') ell, karrScal(i), tauintS(1), tauintS(2)
                !kintresS = kintresS + (tauintS(1)**2)*dkS/q2S**3
             else
                !have to compute primrodal spectrum at right k which is not a constant in tau
                !see above
                write(21,'(I4,5E17.7)') ell, karrScal(i), tauintS(1), tauintS(2), tauintS(3), tauintS(4)
             endif

          enddo
          !write(21,*) ell, 3.23E12*ell*(ell+1.d0)*kintresS
       enddo
       close(21)
       deallocate(SourceFScal)
       deallocate(SourcDatScal)
       deallocate(karrScal)
       deallocate(tauarrScal)
    endif
    if(polarization .and. tensors) then 
       open(unit=28,file=fnPolTLSS, status="replace")
       do s = 1, smaxPolTens
          if(do_log_ell) then
             ell = elltable(s)
          else            
             ell = lmin + lstep*(s-1)
          endif

          !flat sky TT 
          do i = 1, intkmaxPolTens !Integrate over k
             if (i < intkmaxPolTens) then 
                dkT = karrPoltens(i+1) - karrPoltens(i)
             else 
                dkT = karrPolTens(i) - karrPolTens(i-1)
             endif
             tauintT(1:4) = 0.d0
             do j = 1,  inttaumaxPoltens-1 !integration over tau
                if (i .eq. 1) then 
                   dtauT = (tauarrPoltens(2,j+1)-tauarrPoltens(2,j))
                else
                   dtauT = (tauarrPolTens(i,j+1)-tauarrPolTens(i,j))
                endif
                !two different integrals, one with sin, one with cos
                qT = sqrt(karrPolTens(i)**2+(ell/(tauarrPolTens(i,j)-tau0))**2)

	        !using approximation here by replacing tau0->tauR
	        !then also the spectrum should be computed with replacement. 
	        if(AppLSS) then
                   !maresuke thesis page xx
                   !Bfac  = karrPolTens(i)/qT
                   Bfac = 1.0
                   tauintT(1) = tauintT(1) + sin(karrPoltens(i)*(tauarrPolTens(i,j)-tauRT))*SourceFPoltens(i,j,s) & 
                        *dtauT/(tauarrPolTens(i,j)-tau0)**2 * Bfac

                   tauintT(2) = tauintT(2) + cos(karrPolTens(i)*(tauarrPolTens(i,j)-tauRT))*SourceFPoltens(i,j,s) & 
                        *dtauT/(tauarrPoltens(i,j)-tau0)**2 * Bfac
                else 
                   tauintT(1) = tauintT(1) + sin(karrPoltens(i)*(tauarrPoltens(i,j)-tau0))*SourceFPolTens(i,j,s) & 
                        *dtauT/qT**(3.d0)
                   tauintT(2) = tauintT(2) + cos(karrPolTens(i)*(tauarrPolTens(i,j)-tau0))*SourceFPolTens(i,j,s) & 
                        *dtauT/qT**(3.d0) 
                   tauintT(3) = tauintT(3) + sin(karrPolTens(i)*(tauarrPoltens(i,j)-tau0))*SourceFPolTens(i,j,s) & 
                        *dtauT/(tauarrPoltens(i,j)-tau0)**2
                   tauintT(4) = tauintT(4) + cos(karrPolTens(i)*(tauarrPoltens(i,j)-tau0))*SourceFPolTens(i,j,s) & 
                        *dtauT/(tauarrPolTens(i,j)-tau0)**2
                endif


             enddo

             if (AppLSS) then
                if(CLASS) then
                   !have to divide by 4 (the S is 4 times too large from CLASS compared to Seljak/Zaldarriaga)
                   write(28,'(I4,3E17.7)') ell, karrPolTens(i), tauintT(1)/4., tauintT(2)/4.
                else
                   write(28,'(I4,3E17.7)') ell, karrPolTens(i), tauintT(1), tauintT(2)
                endif
                
             else
                !have to compute primordial spectrum at right k which is not a constant in tau
                !see above
                 if(CLASS) then 
                    write(28,'(I4,5E17.7)') ell, karrPolTens(i),  tauintT(1:4)/4.
                 else
                    write(28,'(I4,5E17.7)') ell, karrPolTens(i), tauintT(1:4)
                 endif
                 
             endif

          enddo

       enddo
       close(28)
       deallocate(SourceFPolTens)
       deallocate(SourcDatPolTens)
       deallocate(karrPoltens)
       deallocate(tauarrPoltens)

    endif
    if(polarization .and. scalars) then 
       open(unit=28,file=fnPolSLSS, status="replace")
       do s = 1, smaxPolScal
          if(do_log_ell) then
             ell = elltable(s)
          else            
             ell = lmin + lstep*(s-1)
          endif

          !flat sky TT 
          do i = 1, intkmaxPolScal !Integrate over k
             if (i < intkmaxPolScal) then 
                dkT = karrPolScal(i+1) - karrPolScal(i)
             else 
                dkT = karrPolScal(i) - karrPolScal(i-1)
             endif
             tauintT(1:4) = 0.d0
             do j = 1,  inttaumaxPolScal-1 !integration over tau
                if (i .eq. 1) then 
                   dtauT = (tauarrPolScal(2,j+1)-tauarrPolScal(2,j))
                else
                   dtauT = (tauarrPolScal(i,j+1)-tauarrPolScal(i,j))
                endif
                !two different integrals, one with sin, one with cos
                qT = sqrt(karrPolScal(i)**2+(ell/(tauarrPolScal(i,j)-tau0))**2)

	        !using approximation here by replacing tau0->tauR
	        !then also the spectrum should be computed with replacement. 
	        if(AppLSS) then
                   tauintT(1) = tauintT(1) + sin(karrPolScal(i)*(tauarrPolScal(i,j)-tauRT))*SourceFPolScal(i,j,s) & 
                        *dtauT/(tauarrPolScal(i,j)-tau0)**2 !4 * (ell**2/qT**2) 
                   tauintT(2) = tauintT(2) + cos(karrPolScal(i)*(tauarrPolScal(i,j)-tauRT))*SourceFPolScal(i,j,s) & 
                        *dtauT/(tauarrPolScal(i,j)-tau0)**2 !4 * (ell**2/qT**2)
                else

                   tauintT(1) = tauintT(1) + sin(karrPolScal(i)*(tauarrPolScal(i,j)-tau0))*SourceFPolScal(i,j,s) & 
                        *dtauT/qT**(3.d0)
                   tauintT(2) = tauintT(2) + cos(karrPolScal(i)*(tauarrPolScal(i,j)-tau0))*SourceFPolScal(i,j,s) & 
                        *dtauT/qT**(3.d0) 
                   tauintT(3) = tauintT(3) + sin(karrPolScal(i)*(tauarrPoltens(i,j)-tau0))*SourceFPolScal(i,j,s) & 
                        *dtauT/(tauarrPolScal(i,j)-tau0)**2
                   tauintT(4) = tauintT(4) + cos(karrPolScal(i)*(tauarrPoltens(i,j)-tau0))*SourceFPolScal(i,j,s) & 
                        *dtauT/(tauarrPolScal(i,j)-tau0)**2
                endif


             enddo

             if (AppLSS) then
                if(CLASS) then
                   !see CLASS notes:
                   write(28,'(I4,3E17.7)') ell, karrPolScal(i), sqrt(3./8.)*tauintT(1), sqrt(3./8.)*tauintT(2)
                else
                   write(28,'(I4,3E17.7)') ell, karrPolScal(i), tauintT(1), tauintT(2)
                endif
                
             else
                !have to compute primordial spectrum at right k which is not a constant in tau
                !see above
                write(28,'(I4,5E17.7)') ell, karrPolScal(i), tauintT(1), tauintT(2), tauintT(3), tauintT(4)
             endif

          enddo

       enddo
       close(28)
       deallocate(SourceFPolScal)
       deallocate(SourcDatPolScal)
       deallocate(karrPolScal)
       deallocate(tauarrPolScal)

    endif
    write(*,*) ""


  end subroutine Compute_TransferFlat
end module FlatSkyTranfer
