!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!module computes flat sky transfer functions for B (from tensors) and T (from scalars)!!!!
!Can be modified to compute other transferf (with same input from CAMB) D. Meerburg 2015!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module FlatSkyTranfer

  use settings
  use interpolation

  implicit none


contains 

  subroutine Compute_TransferFlat

    integer     :: i, imax = 100000000, j, n, eof !imax has to be some large number to loop out the file
    real(dl),allocatable    :: SourcDatS(:,:,:) 
    real(dl),allocatable    :: SourcDatT(:,:,:) 
    real(dl),allocatable    :: karrS(:), karrT(:) !number of tau samples for each k
    real(dl),allocatable    :: tauarrS(:,:), tauarrT(:,:)
    real(dl)    :: k, tau, S1, S2, S3
    real(dl)    :: tempk
    integer     :: intkmaxS, intkmaxT, inttaumaxT, inttaumaxS

    real(dl)    :: dkS, dkT, dtauS, dtauT, tauintS(4), tauintT(4), kintresS, kintresT  
    real(dl)    :: qS, qT, q2S, q2T, qS1, qS2 
    real(dl),allocatable :: sp1s(:), sp2s(:), sp3s(:) 	
    real(dl),allocatable :: SourceFS(:,:,:), SourceFT(:,:,:)
    real(dl),allocatable :: sp1t(:), sp2t(:), sp3t(:) 
    integer :: ell, s, smaxT, smaxS
    real(dl) :: TTtensFac
    real(dl) :: qSlow, qTlow, klowS, klowT
    character(100) :: fnSLSS,fnTLSS

    real(dl) :: tauIntegrandS(4), tauIntegrandT(4)



    !required for limited stack size MacOsX + openmp
    allocate(SourcDatS(3,2*lmaxS,200))
    allocate(SourcDatT(3,2*lmaxS,2000))
    allocate(karrS(2*lmaxS))
    allocate(karrT(2*lmaxT))
    !exchanged indices for tau to check for speed. Does not make 
    !a diference so kept it as was for tensors. 
    allocate(tauarrS(200,2*lmaxS))
    allocate(tauarrT(2*lmaxS,2000))

    if (AppLss) then 
       fnSLSS = Trim(homefolder)//Trim(NameTransfer)//Trim(Stxt)
       fnTLSS = Trim(homefolder)//Trim(NameTransfer)//Trim(Ttxt)
    else
       fnSLSS = Trim(homefolder)//Trim(NameTransfer)//Trim('b')//Trim(Stxt)
       fnTLSS = Trim(homefolder)//Trim(NameTransfer)//Trim('b')//Trim(Ttxt)
    endif


    if (scalars) then 
       karrS(1) = 0.d0
       j = 1
       n = 1
       tempk = karrS(1)
       !if bispectrum, use SourcesS_BS.txt
       !else SourcesS.txt or any of the other ones
       fname = Trim(sourcefunctions)//Trim(NameSource)//Trim(Stxt)
       open(unit=21,file=fname, status='old')

       do i = 1, imax

          read(21,'(I5,4E16.7)',iostat=eof)s, k,tau,S1, S2
          !if (j .eq. 10) write(*,*) s, k, tau, S1, S2!, S3 
          if (eof .ne. 0 ) exit !end of file
          if (k > tempk) then
             !numtau(j) = n-1
             !write(*,*) "number of tau:", n-1
             j = j + 1
             n = 1
             !write(*,*) k			
          endif
          !write(*,*) i
          tempk = k
          karrS(j) = k
          tauarrS(n,j) = tau
          !needed for T and E
          SourcDatS(1,j,n) = S1
          SourcDatS(2,j,n) = S2
          !needed for B-modes
          !SourcDat(3,j,n) = S3
          n = n + 1		
       enddo
       intkmaxS = j
       inttaumaxS = n-1
       close(21)
       do i = 1, pstep

          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""
       if (AppLSS) write(*,'(A30)') "Thin Shell Approximation      "
       write(*,'(A30)') "<<<<<<<<<<<<Scalars>>>>>>>>>>>>>"
       write(*,*) ""
       write(*,'(A25,I5)') "Number of tau in scal:     ", inttaumaxS
       write(*,'(A25,I5)') "Number of k in scal:       ", intkmaxS
       write(*,'(A25,I5)') "Delta ell:                 ", lstep
       write(*,'(A25,F5.2)') "kmax:                      ", karrS(intkmaxS)
       write(*,'(A25,F8.2)') "Taumax:                    ", tauarrS(inttaumaxS,2)
       write(*,*) ""
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
    endif
    write(*,*) ""
    if (tensors) then 
       karrT(1) = 0.d0
       j = 1
       n = 1
       tempk = karrT(1)
       !use Sources_BS.txt for bispectrum
       fname = Trim(sourcefunctions)//Trim(NameSource)//Trim(Ttxt)

       open(unit=21,file=fname, status='old')
       do i = 1, imax

          read(21,'(I5,5E16.7)',iostat=eof) s, k,tau,S1, S2, S3
          !if (j .eq. 10) write(*,*) s, k, tau, S1, S2, S3 
          if (eof .ne. 0 ) exit !end of file
          if (k > tempk) then
             !numtau(j) = n-1
             !write(*,*) "number of tau:", n-1
             j = j + 1
             n = 1
             !write(*,*) k			
          endif
          !write(*,*) i
          tempk = k
          karrT(j) = k
          tauarrT(j,n) = tau
          !needed for T and E
          SourcDatT(1,j,n) = S1
          SourcDatT(2,j,n) = S2
          !needed for B-modes
          SourcDatT(3,j,n) = S3
          n = n + 1		
       enddo
       intkmaxT = j
       inttaumaxT = n-1
       close(21)
       !ell = 1000
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""

       if (AppLSS) write(*,'(A30)') "Thin Shell Approximation      "
       write(*,'(A30)') "<<<<<<<<<<<<Tensors>>>>>>>>>>>>>"
       write(*,*) ""
       write(*,'(A25,I5)') "Number of tau in tens:     ", inttaumaxT
       write(*,'(A25,I5)') "Number of k in tens:       ", intkmaxT
       write(*,'(A25,I5)') "Delta ell:                 ", lstep
       write(*,'(A25,F5.2)') "kmax:                      ", karrT(intkmaxT)
       write(*,'(A25,F8.2)') "Taumax:                    ", tauarrT(2,inttaumaxT)
       write(*,*) ""
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
    endif

    !turns out, you can get away with much fewer tau:
    inttaumaxT = 500
    !can not get away with too few k however, especially if you do not use AppLSS = .true.

    smaxT = int(lmaxT/lstep)+1
    smaxS = int(lmaxS/lstep)+1



    !precompute Source function fp
    !this is the slow part:
    if (tensors) then
       !for debugging
       !open(unit=28,file="/mnt/scratch-lustre/meerburg/SourceSplineTp2.txt", status="replace")
       i_progress = int(smaxT/pstep)
       write(*,*) ""
       write(*,'(A40)') "Precomputing Spline Coefficients Tensors"
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""
       allocate(sp1t(1:intkmaxT))
       allocate(sp2t(1:intkmaxT))
       allocate(sp3t(1:intkmaxT))   
       allocate(SourceFT(intkmaxT,inttaumaxT,smaxT))
       !$OMP PARALLEL DO DEFAUlT(SHARED), &
       !$OMP PRIVATE(ell,j,i,s,sp1t,sp2t,sp3t,qT)

       !do ell = lmin, lmaxT, lstep
       do s = 1, smaxT !s = ell
          ell = lmin + lstep*(s-1)
          do j = 1,  inttaumaxT !j = tau

             call spline(karrT(2:intkmaxT),SourcDatT(3,2:intkmaxT,j),sp1t,sp2t,sp3t,intkmaxT-1)


             do i = 1, intkmaxT !loop over k (inverted)
                qT = sqrt(karrT(i)**2+(ell/(tauarrT(i,j)-tau0))**2)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                !This is for B-modes. Change column to 1, if you want T and 2 for E
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                SourceFT(i,j,s) = & 
                     ispline(qT,karrT(2:intkmaxT),SourcDatT(3,2:intkmaxT,j),sp1t,sp2t,sp3t,intkmaxT-1) 	
                !for debugging			
                !if(ell .eq. 90) write(28,*) ell, karrT(i), tauarrT(i,j), SourceFT(i,j,s)
             enddo
          enddo
          !write(*,*) ell
          if(mod(s,i_progress) .eq. 0) write(*,'(a)',advance='no')  '*'

       enddo
       !$OMP END PARALLEL DO 
       deallocate(sp1t)
       deallocate(sp2t)
       deallocate(sp3t)
       !close(28)
    endif

    if (scalars) then
       i_progress = int(smaxS/pstep)
       write(*,*) ""
       write(*,'(A40)') "Precomputing Spline Coefficients Scalars"
       do i = 1, pstep
          write(*,'(a)',advance='no')  '*'
       enddo
       write(*,*) ""
       allocate(sp1s(1:intkmaxS))
       allocate(sp2s(1:intkmaxS))
       allocate(sp3s(1:intkmaxS))
       allocate(SourceFS(intkmaxS,inttaumaxS,smaxS))
       !for debugging
       !open(unit=21,file="/mnt/scratch-lustre/meerburg/SourceSplineS.txt", status="replace")
       !$OMP PARALLEL DO DEFAUlT(SHARED), &
       !$OMP PRIVATE(ell,j,i,s,sp1s,sp2s,sp3s,qS)
       do s = 1, smaxS
          ell = lmin + lstep*(s-1)
          do j = 1,  inttaumaxS 
             call spline(karrS(2:intkmaxS),SourcDatS(1,2:intkmaxS,j),sp1s,sp2s,sp3s,intkmaxS-1)

             do i = 1, intkmaxS !loop over k (inverted)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                !This is for T-modes. Change to columns 2 for E
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                qS = sqrt(karrS(i)**2+(ell/(tauarrS(j,i)-tau0))**2)
                SourceFS(i,j,s) = & 
                     ispline(qS,karrS(2:intkmaxS),SourcDatS(1,2:intkmaxS,j),sp1s,sp2s,sp3s,intkmaxS-1)
                !for debugging		
                !if(ell .eq. 100) write(21,*) ell, karrS(i), tauarrS(i,j), SourceFS(i,j,s)
             enddo
          enddo
          !write(*,*) ell
          if(mod(s,i_progress) .eq. 0) write(*,'(a)',advance='no')  '*'
       enddo
       !$OMP END PARALLEL DO 
       deallocate(sp1s)
       deallocate(sp2s)
       deallocate(sp3s)
       !close(21)
    endif


    if (scalars) then
       open(unit=21,file=fnSLSS, status="replace")
       do s = 1, smaxS !loop over ell
          ell = lmin + lstep*(s-1)
          kintresS = 0.d0
          !flat sky TT 
          do i = 1, intkmaxS !Integrate over k
             if (i < intkmaxS) then 
                dkS = karrS(i+1) - karrS(i)
             else 
                dkS = karrS(i) - karrS(i-1)
             endif
             tauintS(1:4) = 0.d0
             do j = 1,  inttaumaxS-1 !integration over tau
                if (i .eq. 1) then 
                   dtauS = (tauarrS(j+1,2)-tauarrS(j,2))
                else
                   dtauS = (tauarrS(j+1,i)-tauarrS(j,i))
                endif
                !dtauS = tauarrS(j,i)
                !two different integrals, one with sin, one with cos

	        !using approximation here by replacing tau0->tauR
	        !then also the spectrum should be computed with replacement. 
	        if(AppLSS) then
                   TauIntegrandS(1) = (sin(karrS(i)*(tauarrS(j,i)-tauRS))*SourceFS(i,j,s)/(tauarrS(j,i)-tau0)**2 + &
                        sin(karrS(i)*(tauarrS(j+1,i)-tauRS))*SourceFS(i,j+1,s)/(tauarrS(j+1,i)-tau0)**2)/2.0d0
                   tauintS(1) = tauintS(1) + TauIntegrandS(1)*dtauS
                   TauIntegrandS(2) = (cos(karrS(i)*(tauarrS(j,i)-tauRS))*SourceFS(i,j,s)/(tauarrS(j,i)-tau0)**2 + &
                        cos(karrS(i)*(tauarrS(j+1,i)-tauRS))*SourceFS(i,j+1,s)/(tauarrS(j+1,i)-tau0)**2)/2.0d0
                   tauintS(2) = tauintS(2) + TauIntegrandS(2)*dtauS
                   !if (karrS(i) .eq. 0) write(*,*) karrS(i), tauintS(2), SourceFS(i,j,s), TauIntegrandS(2) 
                else
                   !we need to compute the primordial power spectrum with k-par = l^2/D^2
                   qS1 = sqrt(karrS(i)**2+(ell/(tauarrS(j,i)-tau0))**2)
                   qS2 = sqrt(karrS(i)**2+(ell/(tauarrS(j+1,i)-tau0))**2)
                   !multiplied with D_1^2 and divided by k^3
                   TauIntegrandS(1) = (sin(karrS(i)*(tauarrS(j,i)-tauRS))*SourceFS(i,j,s)/qs1**3 + &
                        sin(karrS(i)*(tauarrS(j+1,i)-tauRS))*SourceFS(i,j+1,s)/qs2**3)/2.0d0
                   tauintS(1) = tauintS(1) + TauIntegrandS(1)*dtauS
                   TauIntegrandS(2) = (cos(karrS(i)*(tauarrS(j,i)-tauRS))*SourceFS(i,j,s)/qs1**3 + &
                        cos(karrS(i)*(tauarrS(j+1,i)-tauRS))*SourceFS(i,j+1,s)/qs2**3)/2.0d0
                   tauintS(2) = tauintS(2) + TauIntegrandS(2)*dtauS

                   !same as AppLSS
                   TauIntegrandS(3) = (sin(karrS(i)*(tauarrS(j,i)-tauRS))*SourceFS(i,j,s)/(tauarrS(j,i)-tau0)**2 + &
                        sin(karrS(i)*(tauarrS(j+1,i)-tauRS))*SourceFS(i,j+1,s)/(tauarrS(j+1,i)-tau0)**2)/2.0d0
                   tauintS(3) = tauintS(3) + TauIntegrandS(3)*dtauS
                   TauIntegrandS(4) = (cos(karrS(i)*(tauarrS(j,i)-tauRS))*SourceFS(i,j,s)/(tauarrS(j,i)-tau0)**2 + &
                        cos(karrS(i)*(tauarrS(j+1,i)-tauRS))*SourceFS(i,j+1,s)/(tauarrS(j+1,i)-tau0)**2)/2.0d0
                   tauintS(4) = tauintS(4) + TauIntegrandS(4)*dtauS
                endif

             enddo


             if (AppLSS) then 
                write(21,'(I4,3E17.7)') ell, karrS(i), tauintS(1), tauintS(2)
                !kintresS = kintresS + (tauintS(1)**2)*dkS/q2S**3
             else
                !have to compute primrodal spectrum at right k which is not a constant in tau
                !see above
                write(21,'(I4,5E17.7)') ell, karrS(i), tauintS(1), tauintS(2), tauintS(3), tauintS(4)
             endif

          enddo
          !write(21,*) ell, 3.23E12*ell*(ell+1.d0)*kintresS
       enddo
       close(21)
       deallocate(SourceFS)
    endif
    if(tensors) then 
       open(unit=28,file=fnTLSS, status="replace")
       do s = 1, smaxT   
          ell = lmin + lstep*(s-1)
          kintresT = 0.d0

          !flat sky TT 
          do i = 1, intkmaxT !Integrate over k
             if (i < intkmaxT) then 
                dkT = karrT(i+1) - karrT(i)
             else 
                dkT = karrT(i) - karrT(i-1)
             endif
             tauintT(1:4) = 0.d0
             do j = 1,  inttaumaxT-1 !integration over tau
                if (i .eq. 1) then 
                   dtauT = (tauarrT(2,j+1)-tauarrT(2,j))
                else
                   dtauT = (tauarrT(i,j+1)-tauarrT(i,j))
                endif
                !two different integrals, one with sin, one with cos
                qT = sqrt(karrT(i)**2+(ell/(tauarrT(i,j)-tau0))**2)

	        !using approximation here by replacing tau0->tauR
	        !then also the spectrum should be computed with replacement. 
	        if(AppLSS) then
                   tauintT(1) = tauintT(1) + sin(karrT(i)*(tauarrT(i,j)-tauRT))*SourceFT(i,j,s) & 
                        *dtauT/(tauarrT(i,j)-tau0)**2
                   tauintT(2) = tauintT(2) + cos(karrT(i)*(tauarrT(i,j)-tauRT))*SourceFT(i,j,s) & 
                        *dtauT/(tauarrT(i,j)-tau0)**2
                else
                   !			tauintT(1) = tauintT(1) + karrT(i)*sin(karrT(i)*(tauarrT(i,j)-tau0))*SourceFT(i,j,s) & 
                   !								*dtauT/qT**(4.d0)
                   !			tauintT(2) = tauintT(2) + karrT(i)*cos(karrT(i)*(tauarrT(i,j)-tau0))*SourceFT(i,j,s) & 
                   !								*dtauT/qT**(4.d0) 
                   !			tauintT(3) = tauintT(3) + karrT(i)*sin(karrT(i)*(tauarrT(i,j)-tau0))*SourceFT(i,j,s) & 
                   !								*dtauT/(tauarrT(i,j)-tau0)**2/qT
                   !			tauintT(4) = tauintT(4) + karrT(i)*cos(karrT(i)*(tauarrT(i,j)-tau0))*SourceFT(i,j,s) & 
                   !								*dtauT/(tauarrT(i,j)-tau0)**2/qT
                   !so I am not multiplying by k_z/k since it leads to weird results. We can discuss. Maybe
                   !I am missing something. 
                   tauintT(1) = tauintT(1) + sin(karrT(i)*(tauarrT(i,j)-tau0))*SourceFT(i,j,s) & 
                        *dtauT/qT**(3.d0)
                   tauintT(2) = tauintT(2) + cos(karrT(i)*(tauarrT(i,j)-tau0))*SourceFT(i,j,s) & 
                        *dtauT/qT**(3.d0) 
                   tauintT(3) = tauintT(3) + sin(karrT(i)*(tauarrT(i,j)-tau0))*SourceFT(i,j,s) & 
                        *dtauT/(tauarrT(i,j)-tau0)**2
                   tauintT(4) = tauintT(4) + cos(karrT(i)*(tauarrT(i,j)-tau0))*SourceFT(i,j,s) & 
                        *dtauT/(tauarrT(i,j)-tau0)**2
                endif


             enddo

             if (AppLSS) then 
                write(28,'(I4,3E17.7)') ell, karrT(i), tauintT(1), tauintT(2)
             else
                !have to compute primordial spectrum at right k which is not a constant in tau
                !see above
                write(28,'(I4,5E17.7)') ell, karrT(i), tauintT(1), tauintT(2), tauintT(3), tauintT(4)
             endif

          enddo

       enddo
       close(28)
       deallocate(SourceFT)
    endif
    write(*,*) ""
    deallocate(SourcDatS)
    deallocate(SourcDatT)
    deallocate(karrS)
    deallocate(karrT)
    deallocate(tauarrS)
    deallocate(tauarrT)

  end subroutine Compute_TransferFlat
end module FlatSkyTranfer
