program driver
  use FlatSkyTranfer
  use CMBFlat
  use modules

  !Currently all computed spectra and transfer functions are 
  !dumped. One can add routines to check if certain parts are available
  !or not dump in case you run all routines at once. 

  !To get started, you first have to generate the source functions. This 
  !can be done with the modified version of CAMB. 

  !The name of the source functions has to be adjusted in camb.f90. 
  !search for !pdm to get to the correct line in the file

  !Then in settings.f90 of this code you set the location and name
  !of the source files. Once that has been done you are all set. 

  !You can first run Compute_TransferFlat to obtain the transfer 
  !functions. In settings you can select if you want tensors/scalars

  !There is a flag to either run a thin shell approximation or not 
  !The bispectra can only be computed in the thin shell approximation. 
  !the C_l can do both. 

  !If you compute C_l's you can compare them directly to the one from 
  !CAMB that you generated when you made the source functions

  !the correct files are xxx_scalCls.dat column 1:2 for TT
  !and xxx_tensCls column 1:4 for BB. When I last checked they agreed
  !all the way down to \ell = 10. 

  !See comments in the FlatTransfer.f90 if you want to generate e.g.
  !EE from scalars or tensors (you have to change index of the source
  !function by hand; have never checked this, so be carefull, but should 
  !be straightforward).

  !the bispectra can be computed both for just a slice. See flag below. 
  !the TTT spectra by default is set to the diagonal. The BTT one is 
  !usually set to something else (since the diognal should be zero). 

  !when you set diag_only to false, you compute the full bispectrum. 
  !this is slow. It can take up to a few hours on 16 cores (multi threathed)

  !For testing I suggest you lower kmax and ellmax. I added these as options below. 
  !Make ellmax integer multiples of the number of threads (since it is openmp in l)
  !so e.g 64, 128. Delta ell is set to 5 (can do whatever you want, but slows down 
  !code substabtially). And recall that you have to generate the transfer functions
  !with the same delta ell. 

  !Make sure you do not recompute the transfer functions everytime. Once the transfer
  !functions, comment it out below. As long as you dont change the naming, it will
  !find the transfer functions that you pre-computed earlier. 

  implicit none
  logical :: want_SN
  logical :: diag_only
  integer :: testlmax, testkmax
  integer :: shape
  if (do_log_ell) call set_ell_values
 
  !stop
  !default depends on precomputed transfer functions (ell~2000/5 and k~2000)
  !set to zero if you want to use all.
  !Low kmax will create large errors for high \ell. In the bispectrum this will 
  !lead to large deviations for off diagnoal components (so if you set diag_only to
  !true, it is ok to use only a few hundred k) 
  testlmax = 0
  testkmax = 0

  want_SN = .false.
  diag_only = .false.
!!!COMPUTE TRABNSFER FUNCTION
  !call Compute_TransferFlat 
!!!COMPUTE Cl's. 
  !call Compute_CL_flat
  
!!!COMPUTE TTT bispectrum
  !choose shape
  !local = 1, equileteral  = 2, orthogonal  = 3, flat/enfolded  = 4
  !SFSR = 5, USR = 6
  shape = 1

  !call Compute_Blll_TTT_flat(shape,diag_only, testlmax, testkmax)
  
!!!COMPUTE BTT bispectrum 
  !call Compute_Blll_TTB_flat(shape,diag_only,testlmax, testkmax)

!!!COMPUTE BEE bispectrum 
  !call Compute_Blll_EEB_flat(shape,diag_only,testlmax, testkmax)
  
!!!COMPUTE BTE bispectrum 
  call Compute_Blll_ETB_flat(shape,diag_only,testlmax, testkmax)
  
  !if you already have computed the bispectrum, you can get stand-alone S/N

  !need to know what ellmax integer you used to generate BTT. Will fix all this 
  !later. Also need to know deltaell (and set it in settings.f90)
  !call Get_SN_BTT(testlmax)
  !call Get_SN_TTT(testlmax)
  !call Get_SN_BEE(testlmax)

end program driver
